var path    = require('path')
  , fs      = require('fs')
  , util    = require('util')
  , xml2js  = require('xml2js')
  , _       = require('underscore')
  , upload  = require('jquery-file-upload-middleware')
  , sqlite3 = require('sqlite3').verbose()
  , uuid    = require('node-uuid')
  , date   = require('date-utils');
var isMultiuser = module.parent.exports.set("isMultiuser");


var db = new sqlite3.Database('db.userfile');
exports.datapath = function(req, res){
    var reg = new RegExp(path.delimiter, 'g');
    var url = "file://" + process.cwd().replace(reg, "/") + "/" + getUserdirPath(req, isMultiuser);
    console.log(url);
    res.send(200, url);
};

exports.getFile = function(req, res){
    var username = (isMultiuser) ? req.session.user:"singleuser";
    var filename = req.params.name;
    var filepath = getUserdirPath(req, isMultiuser) + "/" + filename;
    var filesize = fs.statSync(filepath).size;
    console.log(filesize);

    // db = new sqlite3.Database('db.userfile');
    db.serialize(function(){
        var query = "select * from files where username=(?) and filename=(?) and category=(?)";
        db.get(query, [username, filename, "network"], function(err, row){
            console.log(row);
            if(!err && !_.isUndefined(row)){
                var xml = fs.readFileSync(filepath, {encoding: "utf-8"});
                xml = xml.replace(/#!.*batchflow\n/, "");

                var parser = new xml2js.Parser();
                parser.addListener('end', function(result){
                    var data = {
                        type: "network",
                        name: filename,
                        object: xml2data(result)
                    }
                    res.send(200, data);
                    return;
                });
                parser.parseString(xml);
            } else{
                res.send(404, "File not found");
            }
        });
    })
    // db.close();
};

exports.dlFile = function(req, res){
    var username = (isMultiuser)?req.session.user:"singleuser";
    var filename = req.params.name;
    var filetype = req.params.type;
    var isResponded = false;

    // db = new sqlite3.Database('db.userfile');
    db.serialize(function(){
        var query = "select * from files where username=(?) and filename=(?) and category=(?)";
        console.log("username", username, filename, filetype);
        db.get(query, [username, filename, filetype], function(err, row){
            if(!err && !_.isUndefined(row)){
                console.log("file found", getUserdirPath(req, isMultiuser) + "/" + row.filename);
                res.download(getUserdirPath(req, isMultiuser) + "/" + row.filename);
                isResponded = true;
            }
            // for result
            if(!isResponded){
                var path = getUserdirPath(req, isMultiuser) + "/" + filename;
                if(fs.existsSync(path)){
                    res.download(path);
                } else {
                    res.send(404, "File not found");
                }
            }
        });
    })
    // db.close();
};

var isNotDir = function(filepath){return !isDir(filepath);};
var isDir = function( filepath ) {
    if( /\/$/.test( filepath ) ) {
        filepath = filepath.substr( 0, filepath.lastIndexOf('/')-1 );
    }
    return (fs.existsSync( filepath ) && fs.statSync( filepath ).isDirectory());
};
exports.getAllFiles = function(req, res){
    if(isMultiuser && _.isUndefined(req.session.user)){
        console.log("redirect from getAllFiles: isMultiuser=",
                    isMultiuser, ", req.session.user=", req.session.user);
        res.redirect("/login");
    }
    var username = (isMultiuser)?req.session.user:"singleuser";
    // db = new sqlite3.Database('db.userfile');
    db.serialize(function(){
        var query = "select * from files where username=(?)";
        db.all(query, [username], function(err, row){
            var result = [];
            if(row){
                // network/tf/audio category
                for(var i=0; i<row.length; i++){
                    // delete if not exist
                    if(fs.existsSync(getUserdirPath(req, isMultiuser) + "/" + row[i].filename))
                        result.push({name: row[i].filename, type: row[i].category});
                    else
                        db.run("delete from files where username = (?) and filename = (?)",
                               [username, row[i].filename]);
                }

                // results category
                var existingFileNames = _.pluck(result, "name");
                console.log("getUserdirPath(req, isMultiuser)=", getUserdirPath(req, isMultiuser));
                _.chain(fs.readdirSync(getUserdirPath(req, isMultiuser)))
                 .filter(isNotDir)
                 .filter(function(f){return _.indexOf(existingFileNames, f) === -1;})
                 .each(function(f){result.push({name: f, type:"unknown"});})
                 .value();
            }
            // sort by path
            result = _.sortBy(result, function(r){return r.type + r.name;});
            res.send(200, result);
        });
    })
    // db.close();
};
// { id: 'e90e9351-459f-11e3-9ee6-69c23c26711c',
//   username: 'singleuser',
//   filename: 'practice1-2.n',
//   category: 'network',
//   filesize: 0,
//   date: '2013-11-05 07:24:46' }
exports.deleteFile = function(req, res){
    var username = (isMultiuser)?req.session.user:"singleuser";
    var filename = req.params.name;

    // db = new sqlite3.Database('db.userfile');
    db.serialize(function(){
        db.run("delete from files where username=(?) and filename = (?)",
               [username, filename]);
    })
    // db.close();

    var path = getUserdirPath(req, isMultiuser) + "/" + req.params.name;
    if(fs.existsSync(path))
        fs.unlink(path);
    res.send(200, "file deleted");
};

exports.saveNetwork = function (req, res, next) {
    var username = (isMultiuser)?req.session.user:"singleuser";
    var filename = req.params.name;
    var type = "network";

    if(_.isUndefined(req.body.sheets)){
        res.send(400, "Save error: network does not have sheets.");
        return;
    }
    if(!_.contains(_.keys(req.body.sheets), "MAIN")){
        res.send(400, "Save error: network does not have MAIN sheet.");
        return;
    }

    db.get("select * from files where username = (?) and filename = (?)",
           [username, filename], 
        function(err, row){
            if(_.isUndefined(row)){
                var stmt = db.prepare("insert into files values (?, ?, ?, ?, ?, ?)");
                stmt.run([uuid.v1(),
                          username, filename, type,
                          0,
                          (new Date()).toFormat("YYYY-MM-DD HH24:MI:SS")]);
                stmt.finalize();
            }
        }
    );

    saveNetwork(req.params.name, data2xml(req.body), req);
    res.send(200);
}

exports.addPackage = function(req, res, next){
    var username = (isMultiuser)?req.session.user:"singleuser";
    var filename = req.params.name;
    var path = module.parent.exports.set("dirname")+ "/" + getUserdirPath(req, isMultiuser)
                  + "_package/";

    console.log("filename:", filename);
    if(filename.slice(-4) !== ".def" && filename.slice(-4) !== ".tlb") {
        res.send(400, "Extensions must be .def or .tlb");
        return;
    }

    if(fs.existsSync(path + filename)) fs.unlinkSync(path + filename);
    upload.fileHandler({
      uploadDir: function(){return path}, 
      uploadUrl: function(){return "/package/" + filename;},
      tmpDir   : "./tmp"
    })(req, res, next);
}

exports.delPackage = function(req, res, next){
    var username = (isMultiuser)?req.session.user:"singleuser";
    var pkgname = req.params.pkgname;
    var path = module.parent.exports.set("dirname")+ "/" + getUserdirPath(req, isMultiuser)
                  + "_package/";

    if(fs.existsSync(path + pkgname + ".def")) fs.unlinkSync(path + pkgname + ".def");
    if(fs.existsSync(path + pkgname + ".tlb")) fs.unlinkSync(path + pkgname + ".tlb");
    res.send(200);
}

exports.existingPackage = function(req, res){
    var username = (isMultiuser)?req.session.user:"singleuser";
    var path = module.parent.exports.set("dirname")+ "/" + getUserdirPath(req, isMultiuser)
                  + "_package/";

    var pkgs = {};
    var files = _.chain(fs.readdirSync(path))
                 .filter(function(f){
                    return f.slice(-4) === ".tlb" || f.slice(-4) === ".def";})
                 .value();
    for(var i=0; i<files.length; i++){
        var parsed = files[i].split(".");
        console.log(parsed, pkgs, _.keys(pkgs));
        if(_.indexOf(_.keys(pkgs), parsed[0]) === -1){
            pkgs[parsed[0]] = [parsed[1]];
        } else {
            pkgs[parsed[0]].push(parsed[1]);
        }
    }

    res.json(pkgs);
}

exports.addFile = function (req, res, next) {
    var username = (isMultiuser)?req.session.user:"singleuser";
    var type = req.params.type;
    var filename = req.params.name;
    var size = req.params.size || -1;

    var path = module.parent.exports.set("dirname")+ "/" + getUserdirPath(req, isMultiuser)
                  + "/" + filename;

    console.log("added file ", path)
    if(fs.existsSync(path)) fs.unlinkSync(path);
    upload.fileHandler({
      uploadDir: function(){return module.parent.exports.set("dirname")+ "/" + getUserdirPath(req, isMultiuser)}, 
      uploadUrl: function(){return "/file/" + type;},
      tmpDir   : "./tmp"
    })(req, res, next);
    db.get("select * from files where username = (?) and filename = (?)",
           [username, filename], 
        function(err, row){
            if(_.isUndefined(row)){
                db.run("insert into files values (?, ?, ?, ?, ?, ?)",
                       [uuid.v1(),
                        username, filename, type,
                        size,
                        (new Date()).toFormat("YYYY-MM-DD HH24:MI:SS")]);
            }
        }
    );
};

function saveNetwork(name, content, req){
    fs.writeFile(getUserdirPath(req, isMultiuser) + "/" + name, content);
}

function data2xml(network){
    var result = [];

    // templates for xml conersion
    var tmplSheet   = '  <Network type="<%= type %>" name="<%= name %>">';
    var tmplNode    = '    <Node name="<%= id %>" type="<%= name %>" x="<%= pos[0] %>" y="<%= pos[1] %>">';
    var tmplNodeOut = '    <NetOutput name="<%= name %>" node="<%= nodeId %>" terminal="<%= termName %>" object_type="<%- object_type %>" description="<%- description %>"/>';
    var tmplNodeIn  = '    <NetInput name="<%= name %>" node="<%= nodeId %>" terminal="<%= termName %>" object_type="<%- object_type %>" description="<%- description %>"/>';
    var tmplNodeCond= '    <NetCondition name="<%= name %>" node="<%= nodeId %>" terminal="<%= termName %>"/>';
    var tmplParam   = '      <Parameter name="<%= name %>" type="<%- type %>" value="<% if(type==="object"){ %><%- value %><% } else { %><%= value %><% } %>" description="<%- description %>"/>';
    var tmplLink    = '    <Link from="<%= fromNodeId %>" output="<%= fromTermName %>" to="<%= toNodeId %>" input="<%= toTermName %>"/>';

    // headers
    result.push('#!/usr/bin/env batchflow');
    result.push('<?xml version="1.0"?>');
    result.push("<Document>");

    // generate for each sheets
    var nodesPerSheet = _.groupBy(network.nodes, "sheet");
    _.each(_.keys(nodesPerSheet), function(sheet){
        result.push(_.template(tmplSheet, network.sheets[sheet]));

        // generate nodes and their parameters
        _.each(_.filter(network.nodes, function(node){return node.sheet == sheet;}),
            function(node){
            result.push(_.template(tmplNode, node));
            _.chain(node.parameters)
             .each(function(parameter){
                parameter.description = parameter.description.replace(/[ \n\r]*$/g, "");
                result.push(_.template(tmplParam, parameter));
             });
            result.push("    </Node>");
        });

        // generate links
        _.each(_.filter(network.links, function(link){return link.sheet == sheet;}),
            function(link){result.push(_.template(tmplLink, link));}
        );

        // generate network I/O/Condition
        _.each(_.filter(network.terminalRoles, function(role){return role.sheet == sheet;}),
            function(role){
                switch(role.role){
                    case "input"    : result.push(_.template(tmplNodeIn  , role));break;
                    case "output"   : result.push(_.template(tmplNodeOut , role));break;
                    case "condition": result.push(_.template(tmplNodeCond, role));break;
                }
            }
        );
        result.push("  </Network>");
     });

    // footers
    result.push("</Document>");
    result.push("");
    return result.join("\n");
}

function xml2data(xml){
    // console.log(util.inspect(xml, depth=10));
    var sheets = [];
    _.each(xml.Document.Network, function(network){
        // console.log(util.inspect(network, depth=10));
        var sheet = {name: network.$.name,
                     type: network.$.type,
                     links: [],
                     nodes: [],
                     roles: [],
                     notes: []};

        // nodes
        _.each(network.Node, function(node){
            // console.log(util.inspect(node));
            var tmpnode = {
                id: node.$.name,
                index: parseInt(node.$.name.split("_").slice(-1), 10),
                name: node.$.type,
                pos: [Math.floor(parseInt(node.$.x, 10) / 10) * 10,
                      Math.floor(parseInt(node.$.y, 10) / 10) * 10],
                parameters: [],
                sheet: sheet.name
            };

            _.each(node.Parameter, function(param){
                // console.log(util.inspect(param));
                tmpnode.parameters.push({
                    name: param.$.name,
                    type: param.$.type,
                    list: [],
                    value: param.$.value,
                    description: param.$.description
                });
            });
            sheet.nodes.push(tmpnode);
        });

        // links
        _.each(network.Link, function(link){
            // console.log(util.inspect(link));
            var tmplink = {
                fromNodeId: link.$.from,
                fromTermName: link.$.output,
                toNodeId: link.$.to,
                toTermName: link.$.input,
                sheet: sheet.name
            };
            sheet.links.push(tmplink);
        });

        // terminalRoles
        _.each(network.NetInput, function(term){
            // console.log(term);
            sheet.roles.push({
                name       : term.$.name,
                nodeId     : term.$.node,
                description: term.$.description,
                object_type: term.$.object_type,
                role       : "input",
                sheet      : sheet.name,
                termName   : term.$.terminal
            });
        });

        _.each(network.NetOutput, function(term){
            // console.log(term);
            sheet.roles.push({
                name       : term.$.name,
                nodeId     : term.$.node,
                description: term.$.description,
                object_type: term.$.object_type,
                role       : "output",
                sheet      : sheet.name,
                termName   : term.$.terminal
            });
        });

        _.each(network.NetCondition, function(term){
            // console.log(term);
            sheet.roles.push({
                name       : term.$.name,
                nodeId     : term.$.node,
                role       : "condition",
                sheet      : sheet.name,
                termName   : term.$.terminal
            });
        });

        // Node
        _.each(network.Note, function(note){
            // console.log(util.inspect(note));
            var tmpnote = {
                pos: [parseInt(note.$.x, 10), parseInt(note.$.y, 10)],
                visible: parseInt(note.$.visible, 10),
                text: note.$.text
            };
            sheet.notes.push(tmpnote);
        });
        sheets.push(sheet);
    });

    // make inputs for dynamic nodes
    var dynamicNodeNames = _.chain(sheets)
                            .pluck("name")
                            .reject(function(n){return n === "MAIN";})
                            .value();
    var dynamicNodes = _.chain(sheets)
                        .pluck("nodes")
                        .map(function(nodes){
                            return _.filter(nodes, function(n){
                                return dynamicNodeNames.indexOf(n.name) !== -1; });
                        })
                        .flatten()
                        .value();
    _.each(dynamicNodes, function(dyn){
        var sheet = _.find(sheets, function(s){return s.name === dyn.name;});
        dyn.inputs = _.chain(sheet.roles)
                      .filter(function(t){return t.role === "input";})
                      .map(function(t){
                        return {
                            description: t.description,
                            name       : t.name,
                            type       : t.object_type};
                      })
                      .value();
        dyn.outputs = _.chain(sheet.roles)
                      .filter(function(t){return t.role === "output";})
                      .map(function(t){
                        return {
                            description: t.description,
                            name       : t.name,
                            type       : t.object_type};
                      })
                      .value();
    });

    // fix types
    var network = {};
    network.sheets = {};
    _.each(sheets, function(s){network.sheets[s.name] = {name: s.name, type: s.type};});
    network.nodes  = _.chain(sheets).pluck("nodes").flatten().value();
    network.links  = _.chain(sheets).pluck("links").flatten().value();
    network.terminalRoles = _.chain(sheets).pluck("roles").flatten().value();
    network.notes  = _.chain(sheets).pluck("notes").flatten().value();

    // adjust pos for each sheet
    // 1. shift
    _.each(_.keys(network.sheets), function(sheet){
        var common = _.chain(network.nodes)
                      .filter(function(n){return n.sheet === sheet;})
                      .pluck("pos").value();
        var minx = _.chain(common).map(function(p){return p[0];}).min().value();
        var miny = _.chain(common).map(function(p){return p[1];}).min().value();
        _.chain(network.nodes)
         .filter(function(n){return n.sheet === sheet;})
         .each(function(n){
            n.pos[0] -= minx;
            n.pos[1] -= miny;
         });
    });
    // 2. change scale
    var size = [1000, 1000]; // all node positions must be in this size
    _.each(network.nodes, function(n){
        if(n.pos[0] > size[0]) n.pos[0] = size[0];
        if(n.pos[1] > size[1]) n.pos[1] = size[1];
    });

    // re-index all nodes
    if(network.nodes.length > 1) {
        // build new id table.
        var tmp = "___tmp___";
        var counter = {};
        var idtable = {};
        _.each(network.nodes, function(n){
            if(typeof counter[n.name] === "undefined")
                counter[n.name] = 1;
            else
                counter[n.name] += 1;
            idtable[n.sheet + tmp + n.id]
                = "node_" + n.name + "_" + counter[n.name];
        });

        // rename all nodes
        _.each(network.nodes, function(n){
            n.id = idtable[n.sheet + tmp + n.id] || n.id;
        });

        // rename all links
        _.each(network.links, function(l){
            l.fromNodeId = idtable[l.sheet + tmp + l.fromNodeId] || l.fromNodeId;
            l.toNodeId   = idtable[l.sheet + tmp + l.toNodeId] || l.toNodeId;
         });

        // rename all terminalRoles
        _.each(network.terminalRoles, function(r){
            r.nodeId = idtable[r.sheet + tmp + r.nodeId] || r.nodeId;
         });

         // console.log(network.sheets);
    }
    // console.log(network);
    return network;
}
function lock(){
    var lock = path.join(process.cwd(), "lock");
    if(!fs.exists(lock)){
        fs.writeFileSync(lock, "");
        return;
    }

    while(fs.exists(lock)){
        console.log("waiting");
    }
}

function unlock(){
    var lock = path.join(process.cwd(), "lock");
    fs.unlinkSync(lock);
}


function getUserdirPath(req, isMultiuser){
    if(isMultiuser){
        return "userdata/" + req.session.user;
    } else{
        return "userdata/networks";
    }
}


function checkData(obj, type){
    if(type === "network"){
        if(typeof(obj) !== 'string') return "This is not a text file";
        if(obj.indexOf("#!") !== 0) return "This does not starts with #!";
        if(obj.indexOf("<?xml") === -1) return "This not an xml file";
        if(obj.indexOf("batchflow") === -1) return "No word 'bathcflow' found";
        return "";
    }
    else{
        return "unknown data type";
    }
    return "This should not happen";
}
